/*
 * YAFFS: Yet another FFS. A NAND-flash specific file system.
 * yaffs_fs.c
 *
 * Copyright (C) 2002 Aleph One Ltd.
 *   for Toby Churchill Ltd and Brightstar Engineering
 *
 * Created by Charles Manning <charles@aleph1.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This is the file system front-end to YAFFS that hooks it up to
 * the VFS.
 *
 * Special notes: 
 * >> sb->u.generic_sbp points to the yaffs_Device associated with this superblock
 * >> inode->u.generic_ip points to the associated yaffs_Object.
 */


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pagemap.h>
#include <linux/mtd/mtd.h>
#include <linux/interrupt.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <asm/uaccess.h>

#include "yaffs_guts.h"

#ifdef YAFFS_RAM_ENABLED
#include "yaffs_nandemul.h" 
// 2 MB of RAM for emulation
#define YAFFS_RAM_EMULATION_SIZE  0x200000
#endif // YAFFS_RAM_ENABLED

#ifdef YAFFS_MTD_ENABLED
#include <linux/mtd/mtd.h>
#include "yaffs_mtdif.h"
#endif //YAFFS_MTD_ENABLED

#define T(x) printk x



#define yaffs_InodeToObject(iptr) ((yaffs_Object *)((iptr)->u.generic_ip))
#define yaffs_DentryToObject(dptr) yaffs_InodeToObject((dptr)->d_inode)
#define yaffs_SuperToDevice(sb)	((yaffs_Device *)sb->u.generic_sbp)



static ssize_t yaffs_file_read(struct file *f, char *buf, size_t n, loff_t *pos);
static ssize_t yaffs_file_write(struct file *f, const char *buf, size_t n, loff_t *pos);

static int yaffs_sync_object(struct file * file, struct dentry *dentry, int datasync);

static int yaffs_readdir(struct file *f, void *dirent, filldir_t filldir);

static int yaffs_create(struct inode *dir, struct dentry *dentry, int mode);
static struct dentry * yaffs_lookup(struct inode *dir, struct dentry *dentry);
static int yaffs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry);
static int yaffs_unlink(struct inode * dir, struct dentry *dentry);
static int yaffs_symlink(struct inode * dir, struct dentry *dentry, const char * symname);
static int yaffs_mkdir(struct inode * dir, struct dentry * dentry, int mode);
static int yaffs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev);
static int yaffs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry);
static int yaffs_setattr(struct dentry *dentry, struct iattr *attr);

static int yaffs_statfs(struct super_block *sb, struct statfs *buf);
static void yaffs_read_inode (struct inode *inode);
static struct super_block *yaffs_read_super(struct super_block * sb, void * data, int silent);
static void yaffs_put_inode (struct inode *inode);

//static int yaffs_readpage(struct file*,struct page *

static struct address_space_operations yaffs_address_ops = {
//	readpage:		yaffs_readpage,
//	prepare_write:	yaffs_prepare_write,
//	commit_write:	yaffs_commit_write
};


static struct file_operations yaffs_file_operations = {
	read:		yaffs_file_read,
	write:		yaffs_file_write,
//	mmap:		generic_file_mmap,
	fsync:		yaffs_sync_object,
};


static struct inode_operations yaffs_file_inode_operations = {
	setattr:	yaffs_setattr,
};



static struct file_operations yaffs_dir_operations = {
	read:		generic_read_dir,
	readdir:	yaffs_readdir,
	fsync:		yaffs_sync_object,
};

static struct inode_operations yaffs_dir_inode_operations = {
	create:		yaffs_create,
	lookup:		yaffs_lookup,
	link:		yaffs_link,
	unlink:		yaffs_unlink,	
	symlink:	yaffs_symlink,
	mkdir:		yaffs_mkdir,
	rmdir:		yaffs_unlink,
	mknod:		yaffs_mknod,
	rename:		yaffs_rename,
	setattr:	yaffs_setattr,
};

static struct super_operations yaffs_super_ops = {
	statfs:			yaffs_statfs,
	read_inode:		yaffs_read_inode,
	put_inode:		yaffs_put_inode,
//	read_inode:
//	remount_fs:
//	clear_inode:
};


struct inode *yaffs_get_inode(struct super_block *sb, int mode, int dev,yaffs_Object *obj);

/*
 * Lookup is used to find objects in the fs
 */
static struct dentry * yaffs_lookup(struct inode *dir, struct dentry *dentry)
{
	yaffs_Object *obj;
	struct inode *inode;
	
	
	T((KERN_DEBUG"yaffs_lookup for %d:%s\n",yaffs_InodeToObject(dir)->objectId,dentry->d_name.name));
	
	obj = yaffs_FindObjectByName(yaffs_InodeToObject(dir),dentry->d_name.name);
	
	if(obj)
	{
		T((KERN_DEBUG"yaffs_lookup found %d\n",obj->objectId));
		
		inode = yaffs_get_inode(dir->i_sb, obj->st_mode,0,obj);
		
		if(inode)
		{
			T((KERN_DEBUG"yaffs_loookup looks good\n"));
			dget(dentry); // try to solve directory bug
			d_add(dentry,inode);
			return dentry;
		}

	}
	else
	{
		T((KERN_DEBUG"yaffs_lookup not found\n"));
		
	}
	return NULL;
	
}

// For now put inode is just for debugging
static void yaffs_put_inode(struct inode *inode)
{
	T(("yaffs_put_inode: ino %d, count %d\n",(int)inode->i_ino, atomic_read(&inode->i_count)));
}

#ifdef YAFFS_ADDRESS_OPS
static int yaffs_readpage(struct file *file, struct page * page)
{
	T((KERN_DEBUG"yaffs_readpage\n"));

	// TODO
	return 0;
}

static int yaffs_prepare_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{
	T((KERN_DEBUG"yaffs_prepare_write\n"));

	//TODO
	return 0;
}

static int yaffs_commit_write(struct file *file, struct page *page, unsigned offset, unsigned to)
{

	struct inode *inode = page->mapping->host;
	loff_t pos = ((loff_t)page->index << PAGE_CACHE_SHIFT) + to;

	T((KERN_DEBUG"yaffs_commit_write\n"));

	return 0;
}

#endif //YAFFS_ADDRESS_OPS


static void yaffs_FillInodeFromObject(struct inode *inode, yaffs_Object *obj)
{
	if (inode && obj) 
	{
		inode->i_ino = obj->objectId;
		inode->i_mode = obj->st_mode;
		inode->i_uid = obj->st_uid;
		inode->i_gid = obj->st_gid;
		inode->i_blksize = YAFFS_BYTES_PER_CHUNK;
		inode->i_blocks = 0;
		inode->i_rdev = NODEV;
		inode->i_mapping->a_ops = &yaffs_address_ops;
		inode->i_atime = obj->st_atime;
		inode->i_mtime = obj->st_mtime;
		inode->i_ctime = obj->st_ctime;
		inode->i_size = yaffs_GetObjectFileLength(obj);
		inode->i_nlink = yaffs_GetObjectLinkCount(obj);
		
		T((KERN_DEBUG"yaffs_FillInode mode %x uid %d gid %d size %d count %d\n",
				inode->i_mode, inode->i_uid, inode->i_gid, (int)inode->i_size, atomic_read(&inode->i_count)));
		
		switch (obj->st_mode & S_IFMT) 
		{
			default:
			//	init_special_inode(inode, mode, dev);
				break;
			case S_IFREG:	// file		
				inode->i_op = &yaffs_file_inode_operations;
				inode->i_fop = &yaffs_file_operations;
				break;
			case S_IFDIR:	// directory
				inode->i_op = &yaffs_dir_inode_operations;
				inode->i_fop = &yaffs_dir_operations;
				break;
			case S_IFLNK:	// symlink
				inode->i_op = &page_symlink_inode_operations;
				break;
		}
		
		
		inode->u.generic_ip = obj;
		
	}
	else
	{
		T((KERN_DEBUG"yaffs_FileInode invalid parameters\n"));
	}

}

struct inode *yaffs_get_inode(struct super_block *sb, int mode, int dev,yaffs_Object *obj)
{
	struct inode * inode;
	
	T((KERN_DEBUG"yaffs_get_inode for object %d\n",obj->objectId));

	inode = iget(sb,obj->objectId);

	// No need to call this since iget calls it via the read_inode callback
	// yaffs_FillInodeFromObject(inode,obj);

	return inode;
}

static ssize_t yaffs_file_read(struct file *f, char *buf, size_t n, loff_t *pos)
{
	yaffs_Object *obj;
	int nRead,ipos;
	struct inode *inode;
	
	T((KERN_DEBUG"yaffs_file_read\n"));

	obj  = yaffs_DentryToObject(f->f_dentry);
	inode = f->f_dentry->d_inode;
	
	if (*pos < inode->i_size) 
	{
			if (*pos + n > inode->i_size)
			{
				n = inode->i_size - *pos;
			}
	}
	else
	{
		n = 0;
	}
	
	nRead = yaffs_ReadDataFromFile(obj,buf,*pos,n);
	if(nRead > 0)
	{
		f->f_pos += nRead;
	}
	ipos = *pos;
	T((KERN_DEBUG"yaffs_file_read read %d bytes, %d read at %d\n",n,nRead,ipos));
	return nRead;
	
}

static ssize_t yaffs_file_write(struct file *f, const char *buf, size_t n, loff_t *pos)
{
	yaffs_Object *obj;
	int nWritten,ipos;
	struct inode *inode;
	
	obj  = yaffs_DentryToObject(f->f_dentry);
	inode = f->f_dentry->d_inode;
	nWritten = yaffs_WriteDataToFile(obj,buf,*pos,n);
	ipos = *pos;
	T((KERN_DEBUG"yaffs_file_write writing %d bytes, %d written at %d\n",n,nWritten,ipos));
	if(nWritten > 0)
	{
		*pos += nWritten;
		ipos = *pos;
		if(*pos > inode->i_size)
		{
			inode->i_size = *pos;
			T((KERN_DEBUG"yaffs_file_write size updated to %d\n",ipos));
		}
		
	}
	return nWritten;	
}


static int yaffs_readdir(struct file *f, void *dirent, filldir_t filldir)
{
	yaffs_Object *obj;
	struct inode *inode = f->f_dentry->d_inode;
	unsigned long offset, curoffs;
	struct list_head *i;	
	yaffs_Object *l;
	
	char name[YAFFS_MAX_NAME_LENGTH	+1];
		
	obj =  yaffs_DentryToObject(f->f_dentry);
	
	offset = f->f_pos;
	
	T(("yaffs_readdir: starting at %d\n",(int)offset));
	
	if(offset == 0)
	{
        T((KERN_DEBUG"yaffs_readdir: entry . ino %d \n",(int)inode->i_ino));
		if(filldir(dirent,".",1,offset,inode->i_ino,DT_DIR) < 0)
		{
			goto out;
		}
		offset++;
		f->f_pos++;
	}
	if(offset == 1)
	{
		T((KERN_DEBUG"yaffs_readdir: entry .. ino %d \n",(int)f->f_dentry->d_parent->d_inode->i_ino));
		if(filldir(dirent,"..",2,offset,f->f_dentry->d_parent->d_inode->i_ino,DT_DIR) < 0)
		{
			goto out;
		}
		offset++;
		f->f_pos++;
	}
	
	curoffs = 1;
	
	//down(&obj->sem);

	
	list_for_each(i,&obj->variant.directoryVariant.children)
	{
		curoffs++;
		if(curoffs >= offset)
		{		
			l = list_entry(i, yaffs_Object,siblings);
			
			yaffs_GetObjectName(l,name,YAFFS_MAX_NAME_LENGTH+1); 
			T((KERN_DEBUG"yaffs_readdir: %s inode %d\n",name,yaffs_GetObjectInode(l)));
			
			if(filldir(dirent,
					   name,
					   strlen(name),
					   offset,
					   yaffs_GetObjectInode(l),
					   yaffs_GetObjectType(l))
					   < 0)
			{
				goto up_and_out;
			}
			
			offset++;
			f->f_pos++;	   
		}
	}

  up_and_out:

	//up(&obj->sem);
	
  out:
	return 0;
}


/*
 * File creation. Allocate an inode, and we're done..
 */
static int yaffs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev)
{
	struct inode *inode;
	
	yaffs_Object *obj = NULL;
	yaffs_Object *parent = yaffs_InodeToObject(dir);
	
	int error = -ENOSPC;

	if(parent)
	{
		T((KERN_DEBUG"yaffs_mknod: parent object %d type %d\n",
					 parent->objectId,parent->variantType));
	}
	else
	{
		T((KERN_DEBUG"yaffs_mknod: could not get parent object\n"));
		return -EPERM;
	}
	
	T(("yaffs_mknod: making oject for %s, mode %x\n",
				dentry->d_name.name, mode));

	switch (mode & S_IFMT) 
	{
		default:
		
			break;
		case S_IFREG:	// file		
			T((KERN_DEBUG"yaffs_mknod: making file\n"));
			obj = yaffs_MknodFile(parent,dentry->d_name.name,mode,current->uid, current->gid);
			break;
		case S_IFDIR:	// directory
			T((KERN_DEBUG"yaffs_mknod: making directory\n"));
			obj = yaffs_MknodDirectory(parent,dentry->d_name.name,mode,current->uid, current->gid);
			break;
		case S_IFLNK:	// symlink
			T((KERN_DEBUG"yaffs_mknod: making file\n"));
			obj = NULL; // Todo
			break;
	}
	
	if(obj)
	{
		inode = yaffs_get_inode(dir->i_sb, mode, dev, obj);

// did not fix dir bug		if((mode & S_IFMT) == S_IFDIR)	atomic_inc(&inode->i_count);

		d_instantiate(dentry, inode);
		T((KERN_DEBUG"yaffs_mknod created object %d count = %d\n",obj->objectId,atomic_read(&inode->i_count)));
		error = 0;
	}
	else
	{
		T((KERN_DEBUG"yaffs_mknod failed making object\n"));
		error = -ENOMEM;
	}
	

	return error;
}

static int yaffs_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	int retVal;
	T((KERN_DEBUG"yaffs_mkdir\n"));
	retVal =  yaffs_mknod(dir, dentry, mode | S_IFDIR, 0);
#if 0
 // attempt to fix dir bug - didn't work
	if(!retVal)
	{
		dget(dentry);
	}
#endif
	return retVal;
}

static int yaffs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	T((KERN_DEBUG"yaffs_create\n"));
	return yaffs_mknod(dir, dentry, mode | S_IFREG, 0);
}


static int yaffs_unlink(struct inode * dir, struct dentry *dentry)
{
	
	T((KERN_DEBUG"yaffs_unlink\n"));
	
	if(yaffs_Unlink(yaffs_InodeToObject(dir),dentry->d_name.name) == YAFFS_OK)
	{
		return 0;
	}
	else
	{
		return -ENOTEMPTY;
	}
}


/*
 * Link a file..
 */
static int yaffs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry)
{
	struct inode *inode = old_dentry->d_inode;
	
	T((KERN_DEBUG"yaffs_link\n"));
	
	return -EPERM; //Todo
	

	if (S_ISDIR(inode->i_mode))
		return -EPERM;


	return 0;
}


static int yaffs_symlink(struct inode * dir, struct dentry *dentry, const char * symname)
{
	int error;
	
	T((KERN_DEBUG"yaffs_symlink\n"));

	
	return -ENOMEM; //Todo

	error = yaffs_mknod(dir, dentry, S_IFLNK | S_IRWXUGO, 0);
	return error;
}

static int yaffs_sync_object(struct file * file, struct dentry *dentry, int datasync)
{

	T((KERN_DEBUG"yaffs_sync_object\n"));
	yaffs_FlushFile(yaffs_DentryToObject(dentry));
	return 0;
}

/*
 * The VFS layer already does all the dentry stuff for rename.
 */
static int yaffs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry)
{
	
	
	if( yaffs_RenameObject(yaffs_InodeToObject(old_dir),old_dentry->d_name.name,
	                       yaffs_InodeToObject(new_dir),new_dentry->d_name.name) == YAFFS_OK)
	{
		return 0;
	}
	else
	{
		return -ENOTEMPTY;
	}
	

}

static int yaffs_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct inode *inode = dentry->d_inode;
	int error;
	
	T((KERN_DEBUG"yaffs_setattr\n"));
	
	if((error = inode_change_ok(inode,attr)) == 0)
	{
	
		if(yaffs_SetAttributes(yaffs_InodeToObject(inode),attr) == YAFFS_OK)
		{
			error = 0;
		}
		else
		{
			error = -EPERM;
		}

		inode_setattr(inode,attr);
	}
	return error;
}

static int yaffs_statfs(struct super_block *sb, struct statfs *buf)
{
	T((KERN_DEBUG"yaffs_statfs\n"));

	buf->f_type = YAFFS_MAGIC;
	buf->f_bsize = YAFFS_BYTES_PER_CHUNK;
	buf->f_namelen = 255;
	buf->f_blocks = yaffs_SuperToDevice(sb)->nBlocks * YAFFS_CHUNKS_PER_BLOCK;
	buf->f_files = 0;
	buf->f_ffree = 0;
	buf->f_bavail = buf->f_bfree = yaffs_GetNumberOfFreeChunks(yaffs_SuperToDevice(sb));
	return 0;
}

static void yaffs_read_inode (struct inode *inode)
{

	yaffs_Object *obj ; 
	
	T((KERN_DEBUG"yaffs_read_inode for %d\n",(int)inode->i_ino));

	obj  = yaffs_FindObjectByNumber(yaffs_SuperToDevice(inode->i_sb),inode->i_ino);
	
	yaffs_FillInodeFromObject(inode,obj);
}


static struct super_block *yaffs_internal_read_super(int useRam, struct super_block * sb, void * data, int silent)
{
	struct inode * inode;
	struct dentry * root;
	yaffs_Device *dev;
	

	T(("yaffs_read_super: %s\n", useRam ? "RAM" : "MTD"));
	sb->s_blocksize = YAFFS_BYTES_PER_CHUNK;
	sb->s_blocksize_bits = YAFFS_CHUNK_SIZE_SHIFT;
	sb->s_magic = YAFFS_MAGIC;
	sb->s_op = &yaffs_super_ops;
	
	if(!sb)
		printk(KERN_INFO"yaffs: sb is NULL\n");
	else if(!sb->s_dev)
		printk(KERN_INFO"yaffs: sb->s_dev is NULL\n");
	else if(! kdevname(sb->s_dev))
		printk(KERN_INFO"yaffs: kdevname is NULL\n");
	else
		printk(KERN_INFO"yaffs: dev is %d name is \"%s\"\n", sb->s_dev, kdevname(sb->s_dev));

	
	
	if(useRam)
	{

#if YAFFS_RAM_ENABLED
		// Set the yaffs_Device up for ram emulation

		sb->u.generic_sbp =	dev = kmalloc(sizeof(yaffs_Device),GFP_KERNEL);
		if(!dev)
		{
			// Deep shit could not allocate device structure
			printk(KERN_DEBUG"yaffs_read_super: Failed trying to allocate yaffs_Device. Terminating debug.\n");
			return NULL;
		}

		memset(dev,0,sizeof(yaffs_Device));
		dev->genericDevice = NULL; // Not used for RAM emulation.

		dev->nBlocks = YAFFS_RAM_EMULATION_SIZE / (YAFFS_CHUNKS_PER_BLOCK * YAFFS_BYTES_PER_CHUNK);
		dev->startBlock = 1;  // Don't use block 0
		dev->endBlock = dev->nBlocks - 1;

		dev->writeChunkToNAND = nandemul_WriteChunkToNAND;
		dev->readChunkFromNAND = nandemul_ReadChunkFromNAND;
		dev->eraseBlockInNAND = nandemul_EraseBlockInNAND;
		dev->initialiseNAND = nandemul_InitialiseNAND;
#endif

	}
	else
	{	
#ifdef YAFFS_MTD_ENABLED
		struct mtd_info *mtd;
		
		printk(KERN_DEBUG "yaffs: Attempting MTD mount on %u.%u, \"%s\"\n",
		 MAJOR(sb->s_dev),MINOR(sb->s_dev),kdevname(sb->s_dev));
			
		// Hope it's a NAND mtd
		mtd = get_mtd_device(NULL, MINOR(sb->s_dev));
		if (!mtd) 
		{
			printk(KERN_DEBUG "yaffs: MTD device #%u doesn't appear to exist\n", MINOR(sb->s_dev));
			return NULL;
		}
		
		if(mtd->type != MTD_NANDFLASH)
		{
			printk(KERN_DEBUG "yaffs: MTD device is not NAND it's type %d\n", mtd->type);
			return NULL;
		}

		printk(KERN_DEBUG" erase %x\n",mtd->erase);
		printk(KERN_DEBUG" read %x\n",mtd->read);
		printk(KERN_DEBUG" write %x\n",mtd->write);
		printk(KERN_DEBUG" readoob %x\n",mtd->read_oob);
		printk(KERN_DEBUG" writeoob %x\n",mtd->write_oob);
		printk(KERN_DEBUG" oobblock %x\n",mtd->oobblock);
		printk(KERN_DEBUG" oobsize %x\n",mtd->oobsize);


		if(!mtd->erase ||
		   !mtd->read  ||
		   !mtd->write ||
		   !mtd->read_oob ||
		   !mtd->write_oob)
		{
			printk(KERN_DEBUG "yaffs: MTD device does not support required functions\n");
			return NULL;
		}
		
		if(mtd->oobblock != YAFFS_BYTES_PER_CHUNK ||
		   mtd->oobsize != YAFFS_BYTES_PER_SPARE)
		{
			printk(KERN_DEBUG "yaffs: MTD device does not support have the right page sizes\n");
			return NULL;
		}
		   

		// OK, so if we got here, we have an MTD that's NAND and looks 
		// like it has the right capabilities
		// Set the yaffs_Device up for ram emulation

		sb->u.generic_sbp =	dev = kmalloc(sizeof(yaffs_Device),GFP_KERNEL);
		if(!dev)
		{
			// Deep shit could not allocate device structure
			printk(KERN_DEBUG"yaffs_read_super: Failed trying to allocate yaffs_Device. Terminating debug.\n");
			return NULL;
		}

		memset(dev,0,sizeof(yaffs_Device));
		dev->genericDevice = mtd; 

		// Set up the memory size parameters....
		
// NCB		dev->nBlocks = YAFFS_RAM_EMULATION_SIZE / (YAFFS_CHUNKS_PER_BLOCK * YAFFS_BYTES_PER_CHUNK);
		dev->nBlocks = mtd->size / (YAFFS_CHUNKS_PER_BLOCK * YAFFS_BYTES_PER_CHUNK);
		dev->startBlock = 1;  // Don't use block 0
		dev->endBlock = dev->nBlocks - 1;

		// ... and the functions.
		dev->writeChunkToNAND = nandmtd_WriteChunkToNAND;
		dev->readChunkFromNAND = nandmtd_ReadChunkFromNAND;
		dev->eraseBlockInNAND = nandmtd_EraseBlockInNAND;
		dev->initialiseNAND = nandmtd_InitialiseNAND;
		
#endif
	}



	yaffs_GutsInitialise(yaffs_SuperToDevice(sb));
	T(("yaffs_read_super: guts initialised\n"));

	// Create root inode
	inode = yaffs_get_inode(sb, S_IFDIR | 0755, 0,yaffs_Root(yaffs_SuperToDevice(sb)));
	if (!inode)
		return NULL;

	T(("yaffs_read_super: got root inode\n"));
		

	root = d_alloc_root(inode);

	T(("yaffs_read_super: d_alloc_root done\n"));

	if (!root) {
		iput(inode);
		return NULL;
	}
	sb->s_root = root;

	T(("yaffs_read_super: done\n"));
	return sb;
}

#ifdef YAFFS_MTD_ENABLED
static struct super_block *yaffs_read_super(struct super_block * sb, void * data, int silent)
{
	return yaffs_internal_read_super(0,sb,data,silent);
}

static DECLARE_FSTYPE(yaffs_fs_type, "yaffs", yaffs_read_super, FS_REQUIRES_DEV);
#endif

#ifdef YAFFS_RAM_ENABLED

static struct super_block *yaffs_ram_read_super(struct super_block * sb, void * data, int silent)
{
	return yaffs_internal_read_super(1,sb,data,silent);
}

static DECLARE_FSTYPE(yaffs_ram_fs_type, "yaffsram", yaffs_ram_read_super, FS_SINGLE);
#endif // YAFFS_RAM_ENABLED


static struct proc_dir_entry *my_proc_entry;


static int  yaffs_proc_read(
        char *page,
	char **start,
	off_t offset,
	int count,
	int *eof,
	void *data
	)
{

	static char my_buffer[1000];

	if (offset > 0) return 0;

	/* Fill the buffer and get its length */
	sprintf( my_buffer, 
		"YAFFS built:"__DATE__ " "__TIME__"\n"
		
	);

	strcpy(page,my_buffer);
	return strlen(my_buffer);
}

static int __init init_yaffs_fs(void)
{
	int error = 0;
	
	printk(KERN_DEBUG "yaffs " __DATE__ " " __TIME__ " Initialisation\n");
    /* Install the proc_fs entry */
    my_proc_entry = create_proc_read_entry("yaffs",
                                           S_IRUGO | S_IFREG,
					   &proc_root,
					   yaffs_proc_read,
					   NULL);
    if(!my_proc_entry)
    {
       return -ENOMEM;
    }

#ifdef YAFFS_RAM_ENABLED

    error = register_filesystem(&yaffs_ram_fs_type);
    if(error)
    {
    	return error;
    }
#endif //YAFFS_RAM_ENABLED

#ifdef YAFFS_MTD_ENABLED
	error = register_filesystem(&yaffs_fs_type);
	if(error)
	{
#ifdef YAFFS_RAM_ENABLED
		unregister_filesystem(&yaffs_ram_fs_type);
#endif //YAFFS_RAM_ENABLED
	}
#endif // YAFFS_MTD_ENABLED

	return error;
}

static void __exit exit_yaffs_fs(void)
{
	printk(KERN_DEBUG "yaffs " __DATE__ " " __TIME__ " Clean up\n");

    remove_proc_entry("yaffs",&proc_root);
    
#ifdef YAFFS_RAM_ENABLED
// NCB	unregister_filesystem(&yaffs_fs_type);
	unregister_filesystem(&yaffs_ram_fs_type);
#endif
#ifdef YAFFS_MTD_ENABLED
// NCB	unregister_filesystem(&yaffs_ram_fs_type);
	unregister_filesystem(&yaffs_fs_type);
#endif

}

module_init(init_yaffs_fs)
module_exit(exit_yaffs_fs)

MODULE_DESCRIPTION("YAFFS - a NAND specific flash file system");
MODULE_AUTHOR("Charles Manning, Aleph One Ltd., 2002");
MODULE_LICENSE("GPL");

